from .point import Point
from .to_sql_interface import ToSQLInterface


class Polygon(ToSQLInterface):
    def __init__(self, points: list[Point]):
        if len(points) < 3:
            raise ValueError("A polygon must have at least 3 points.")
        self.points = points

    def to_sql(self):
        # the first point as the last point to ensure the polygon is closed.
        points_sql = ", ".join([f"{point.longitude} {point.latitude}" for point in self.points] + [
            f"{self.points[0].longitude} {self.points[0].latitude}"])
        return f"ST_PolygonFromText('POLYGON(({points_sql}))')"

    @staticmethod
    def create_select_stmt(column_name: str):
        return f"ST_AsText({column_name})"

    def __eq__(self, other_polygon):
        if isinstance(other_polygon, Polygon):
            return self.points == other_polygon.points
        return False

    @classmethod
    def from_text(cls, polygon_text):

        # Example input: 'POLYGON((10.1 10.1,10.2 10.2,10.3 10.3,10.1 10.1))'
        # Remove the outer parentheses
        inner_text = polygon_text.replace('POLYGON((', '').replace('))', '')

        # Splitting the inner text into pairs and
        # converting them to Point instances
        pairs = inner_text.split(',')
        # Converting each pair to a Point instance, Exclude the last pair
        points = [Point(float(pair.split()[0]), float(pair.split()[1])) for pair in pairs[:-1]]
        return cls(points)



from .to_sql_interface import ToSQLInterface


class Point(ToSQLInterface):
    def __init__(self, longitude, latitude):
        self.longitude = longitude
        self.latitude = latitude

    def to_sql(self):
        return f"POINT ({self.longitude}, {self.latitude})"

    @staticmethod
    def create_select_stmt(column_name: str):
        return f"ST_X({column_name}), ST_Y({column_name})"

    def __eq__(self, other_point):
        if isinstance(other_point, Point):
            return self.longitude == other_point.longitude and self.latitude == other_point.latitude
        return False



import setuptools

PACKAGE_NAME = "database-mysql-local"
package_dir = PACKAGE_NAME.replace("-", "_")
# TODO: we are migrating from circles_local_database_python to database_mysql_local
#  but in the meantime we want to keep both names
old_name = "circles_local_database_python"

setuptools.setup(
    name=PACKAGE_NAME,  # https://pypi.org/project/database-mysql-local
    version='0.0.230',
    author="Circles",
    author_email="info@circles.life",
    url=f"https://github.com/circles-zone/{PACKAGE_NAME}-python-package",
    packages=[package_dir, old_name],
    package_dir={package_dir: f'{package_dir}/src', old_name: f'{package_dir}/src'},
    package_data={package_dir: ['*.py']},
    long_description="Database MySQL Local",
    long_description_content_type='text/markdown',
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: Other/Proprietary License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        "mysql-connector-python>=8.3.0",  # https://pypi.org/project/mysql-connector-python/
        "url_remote>=0.0.68",  # https://pypi.org/project/url-remote/
        "logger-local>=0.0.102",  # https://pypi.org/project/logger-local/
        "database-infrastructure-local>=0.0.19",  # https://pypi.org/project/database-infrastructure-local/
        "language-remote>=0.0.8",  # https://pypi.org/project/language-local/
        "sql-to-code-local>=0.0.2",  # https://pypi.org/project/sql-to-code-local/
        "python-sdk-remote>=0.0.75"
    ]
)



# TODO: Regenerate this file with sql2code
table_definition = {
    'phone_table': {
        'table_definition_id': 1,
        'entity_id': None,
        'schema': 'phone',
        'github_organization': '',
        'github_repo': '',
        'seq': 0,
        'table_list_id': None,
        'is_created_updated_fields': 1,
        'is_logical_delete_fields': 1,
        'profile_mapping_table_id': 2,
        'is_ml_table': None,
        'table_type': None,
        'is_number_column': 1,
        'is_test_data_column': None,
        'is_main_column': None,
        'is_single_and_multi_value': None,
        'name_of_single_value_column': None,
        'is_mapping_table': None,
        'is_entity': None,
        'is_history': None,
    },
    'phone_profile_table': {
        'table_definition_id': 2,
        'entity_id': None,
        'schema': 'phone_profile',
        'github_organization': '',
        'github_repo': '',
        'seq': 0,
        'table_list_id': None,
        'is_created_updated_fields': 1,
        'is_logical_delete_fields': 1,
        'profile_mapping_table_id': None,
        'is_ml_table': None,
        'table_type': None,
        'is_number_column': 0,
        'is_test_data_column': None,
        'is_main_column': None,
        'is_single_and_multi_value': None,
        'name_of_single_value_column': None,
        'is_mapping_table': None,
        'is_entity': None,
        'is_history': None,
    },
    'email_address_profile_table': {
        'table_definition_id': 3,
        'entity_id': None,
        'schema': 'email_address_profile',
        'github_organization': '',
        'github_repo': '',
        'seq': 0,
        'table_list_id': None,
        'is_created_updated_fields': 1,
        'is_logical_delete_fields': 1,
        'profile_mapping_table_id': None,
        'is_ml_table': None,
        'table_type': None,
        'is_number_column': 0,
        'is_test_data_column': None,
        'is_main_column': None,
        'is_single_and_multi_value': None,
        'name_of_single_value_column': None,
        'is_mapping_table': None,
        'is_entity': None,
        'is_history': None,
    },
    'email_address_table': {
        'table_definition_id': 4,
        'entity_id': None,
        'schema': 'email_address',
        'github_organization': '',
        'github_repo': '',
        'seq': 0,
        'table_list_id': None,
        'is_created_updated_fields': 1,
        'is_logical_delete_fields': 1,
        'profile_mapping_table_id': 3,
        'is_ml_table': None,
        'table_type': None,
        'is_number_column': 1,
        'is_test_data_column': None,
        'is_main_column': None,
        'is_single_and_multi_value': None,
        'name_of_single_value_column': None,
        'is_mapping_table': None,
        'is_entity': None,
        'is_history': None,
    },
    'user_table': {
        'table_definition_id': 5,
        'entity_id': None,
        'schema': 'user',
        'github_organization': '',
        'github_repo': '',
        'seq': 0,
        'table_list_id': None,
        'is_created_updated_fields': 1,
        'is_logical_delete_fields': 1,
        'profile_mapping_table_id': 6,
        'is_ml_table': None,
        'table_type': None,
        'is_number_column': 1,
        'is_test_data_column': None,
        'is_main_column': None,
        'is_single_and_multi_value': None,
        'name_of_single_value_column': None,
        'is_mapping_table': None,
        'is_entity': None,
        'is_history': None,
    },
    'profile_user_table': {
        'table_definition_id': 6,
        'entity_id': None,
        'schema': 'profile_user',
        'github_organization': '',
        'github_repo': '',
        'seq': 0,
        'table_list_id': None,
        'is_created_updated_fields': 1,
        'is_logical_delete_fields': 1,
        'profile_mapping_table_id': None,
        'is_ml_table': None,
        'table_type': None,
        'is_number_column': 0,
        'is_test_data_column': None,
        'is_main_column': None,
        'is_single_and_multi_value': None,
        'name_of_single_value_column': None,
        'is_mapping_table': None,
        'is_entity': None,
        'is_history': None,
    },
    'group_profile_table': {
        'table_definition_id': 7,
        'entity_id': None,
        'schema': 'group_profile',
        'github_organization': '',
        'github_repo': '',
        'seq': 0,
        'table_list_id': None,
        'is_created_updated_fields': 1,
        'is_logical_delete_fields': 1,
        'profile_mapping_table_id': None,
        'is_ml_table': None,
        'table_type': None,
        'is_number_column': 0,
        'is_test_data_column': None,
        'is_main_column': None,
        'is_single_and_multi_value': None,
        'name_of_single_value_column': None,
        'is_mapping_table': None,
        'is_entity': None,
        'is_history': None,
    },
    'group_table': {
        'table_definition_id': 8,
        'entity_id': None,
        'schema': 'group',
        'github_organization': '',
        'github_repo': '',
        'seq': 0,
        'table_list_id': None,
        'is_created_updated_fields': 1,
        'is_logical_delete_fields': 1,
        'profile_mapping_table_id': 7,
        'is_ml_table': 1,
        'table_type': None,
        'is_number_column': 1,
        'is_test_data_column': None,
        'is_main_column': None,
        'is_single_and_multi_value': None,
        'name_of_single_value_column': None,
        'is_mapping_table': None,
        'is_entity': None,
        'is_history': None,
    },
    'group_label_table': {
        'table_definition_id': 9,
        'entity_id': None,
        'schema': 'group_label',
        'github_organization': '',
        'github_repo': '',
        'seq': 0,
        'table_list_id': None,
        'is_created_updated_fields': 1,
        'is_logical_delete_fields': 1,
        'profile_mapping_table_id': None,
        'is_ml_table': None,
        'table_type': None,
        'is_number_column': 0,
        'is_test_data_column': None,
        'is_main_column': None,
        'is_single_and_multi_value': None,
        'name_of_single_value_column': None,
        'is_mapping_table': None,
        'is_entity': None,
        'is_history': None,
    },
    'profile_metics_table': {
        'table_definition_id': 10,
        'entity_id': None,
        'schema': 'profile_metics',
        'github_organization': '',
        'github_repo': '',
        'seq': 0,
        'table_list_id': None,
        'is_created_updated_fields': 1,
        'is_logical_delete_fields': 1,
        'profile_mapping_table_id': None,
        'is_ml_table': None,
        'table_type': None,
        'is_number_column': 0,
        'is_test_data_column': None,
        'is_main_column': None,
        'is_single_and_multi_value': None,
        'name_of_single_value_column': None,
        'is_mapping_table': None,
        'is_entity': None,
        'is_history': None,
    },
    'profile_url_table': {
        'table_definition_id': 11,
        'entity_id': None,
        'schema': 'profile_url',
        'github_organization': '',
        'github_repo': '',
        'seq': 0,
        'table_list_id': None,
        'is_created_updated_fields': 1,
        'is_logical_delete_fields': 1,
        'profile_mapping_table_id': None,
        'is_ml_table': None,
        'table_type': None,
        'is_number_column': 0,
        'is_test_data_column': None,
        'is_main_column': None,
        'is_single_and_multi_value': None,
        'name_of_single_value_column': None,
        'is_mapping_table': None,
        'is_entity': None,
        'is_history': None,
    },
    'user_external_table': {
        'table_definition_id': 12,
        'entity_id': None,
        'schema': 'user_external',
        'github_organization': '',
        'github_repo': '',
        'seq': 0,
        'table_list_id': None,
        'is_created_updated_fields': 1,
        'is_logical_delete_fields': 1,
        'profile_mapping_table_id': None,
        'is_ml_table': None,
        'table_type': None,
        'is_number_column': 1,
        'is_test_data_column': None,
        'is_main_column': None,
        'is_single_and_multi_value': None,
        'name_of_single_value_column': None,
        'is_mapping_table': None,
        'is_entity': None,
        'is_history': None,
    },
    'location_table': {
        'table_definition_id': 13,
        'entity_id': None,
        'schema': 'location',
        'github_organization': '',
        'github_repo': '',
        'seq': 0,
        'table_list_id': None,
        'is_created_updated_fields': 1,
        'is_logical_delete_fields': 1,
        'profile_mapping_table_id': 14,
        'is_ml_table': None,
        'table_type': None,
        'is_number_column': 1,
        'is_test_data_column': None,
        'is_main_column': None,
        'is_single_and_multi_value': None,
        'name_of_single_value_column': None,
        'is_mapping_table': None,
        'is_entity': None,
        'is_history': None,
    },
    'location_profile_table': {
        'table_definition_id': 14,
        'entity_id': None,
        'schema': 'location_profile',
        'github_organization': '',
        'github_repo': '',
        'seq': 0,
        'table_list_id': None,
        'is_created_updated_fields': 1,
        'is_logical_delete_fields': 1,
        'profile_mapping_table_id': None,
        'is_ml_table': None,
        'table_type': None,
        'is_number_column': 0,
        'is_test_data_column': None,
        'is_main_column': None,
        'is_single_and_multi_value': None,
        'name_of_single_value_column': None,
        'is_mapping_table': None,
        'is_entity': None,
        'is_history': None,
    },
    'profile_profile_table': {
        'table_definition_id': 15,
        'entity_id': None,
        'schema': 'profile_profile',
        'github_organization': '',
        'github_repo': '',
        'seq': 0,
        'table_list_id': None,
        'is_created_updated_fields': 1,
        'is_logical_delete_fields': 1,
        'profile_mapping_table_id': None,
        'is_ml_table': None,
        'table_type': None,
        'is_number_column': 0,
        'is_test_data_column': None,
        'is_main_column': None,
        'is_single_and_multi_value': None,
        'name_of_single_value_column': None,
        'is_mapping_table': None,
        'is_entity': None,
        'is_history': None,
    },
    'profile_table': {
        'table_definition_id': 16,
        'entity_id': None,
        'schema': 'profile',
        'github_organization': '',
        'github_repo': '',
        'seq': 0,
        'table_list_id': None,
        'is_created_updated_fields': 1,
        'is_logical_delete_fields': 1,
        'profile_mapping_table_id': None,
        'is_ml_table': 1,
        'table_type': None,
        'is_number_column': 1,
        'is_test_data_column': None,
        'is_main_column': None,
        'is_single_and_multi_value': None,
        'name_of_single_value_column': None,
        'is_mapping_table': None,
        'is_entity': None,
        'is_history': None,
    },
    'person_table': {
        'table_definition_id': 17,
        'entity_id': None,
        'schema': 'person',
        'github_organization': '',
        'github_repo': '',
        'seq': 0,
        'table_list_id': None,
        'is_created_updated_fields': 1,
        'is_logical_delete_fields': 1,
        'profile_mapping_table_id': None,
        'is_ml_table': None,
        'table_type': None,
        'is_number_column': 1,
        'is_test_data_column': None,
        'is_main_column': None,
        'is_single_and_multi_value': 1,
        'name_of_single_value_column': 'main_email_address',
        'is_mapping_table': None,
        'is_entity': None,
        'is_history': None,
    },
    'test_mysql_table': {
        'table_definition_id': 18,
        'entity_id': None,
        'schema': 'test',
        'github_organization': '',
        'github_repo': '',
        'seq': 0,
        'table_list_id': None,
        'is_created_updated_fields': 1,
        'is_logical_delete_fields': 1,
        'profile_mapping_table_id': None,
        'is_ml_table': None,
        'table_type': None,
        'is_number_column': 1,
        'is_test_data_column': None,
        'is_main_column': None,
        'is_single_and_multi_value': None,
        'name_of_single_value_column': None,
        'is_mapping_table': None,
        'is_entity': None,
        'is_history': None,
    },
    'event_table': {
        'table_definition_id': 19,
        'entity_id': None,
        'schema': 'event',
        'github_organization': '',
        'github_repo': '',
        'seq': 0,
        'table_list_id': None,
        'is_created_updated_fields': 1,
        'is_logical_delete_fields': 1,
        'profile_mapping_table_id': None,
        'is_ml_table': None,
        'table_type': None,
        'is_number_column': None,
        'is_test_data_column': None,
        'is_main_column': None,
        'is_single_and_multi_value': None,
        'name_of_single_value_column': None,
        'is_mapping_table': None,
        'is_entity': None,
        'is_history': None,
    },
    'logger_table': {
        'table_definition_id': 20,
        'entity_id': None,
        'schema': 'logger',
        'github_organization': '',
        'github_repo': '',
        'seq': 0,
        'table_list_id': None,
        'is_created_updated_fields': 1,
        'is_logical_delete_fields': 1,
        'profile_mapping_table_id': None,
        'is_ml_table': None,
        'table_type': None,
        'is_number_column': None,
        'is_test_data_column': None,
        'is_main_column': None,
        'is_single_and_multi_value': None,
        'name_of_single_value_column': None,
        'is_mapping_table': None,
        'is_entity': None,
        'is_history': 1,
    },
}



from __future__ import annotations  # Required for type hinting in class methods

import mysql.connector
from logger_local.LoggerLocal import Logger
from logger_local.MetaLogger import MetaLogger
from python_sdk_remote.utilities import get_sql_hostname, get_sql_username, get_sql_password

from .constants import LOGGER_CONNECTOR_CODE_OBJECT
from .cursor import Cursor

logger = Logger.create_logger(object=LOGGER_CONNECTOR_CODE_OBJECT)
connections_pool = {}


# TODO Can we add hostname, IPv4, IPv6, sql statement? Can we have the same message format in all cases?
class Connector(metaclass=MetaLogger, object=LOGGER_CONNECTOR_CODE_OBJECT):
    @staticmethod
    def connect(schema_name: str) -> Connector:
        if (schema_name in connections_pool and
                connections_pool[schema_name] and
                connections_pool[schema_name].connection):
            if connections_pool[schema_name].connection.is_connected():

                return connections_pool[schema_name]
            else:
                # reconnect
                connections_pool[schema_name].connection.reconnect()
                # TODO We should develop retry mechanism to support password rotation and small network issues.
                if connections_pool[schema_name].connection.is_connected():

                    return connections_pool[schema_name]
                else:
                    connector = Connector(schema_name)
                    logger.warning("Reconnect failed, returning a new connection",
                                   object={'connections_pool': str(connections_pool[schema_name])})

                    return connector
        else:
            connector = Connector(schema_name)

            return connector

    def __init__(self, schema_name: str,
                 host: str = get_sql_hostname(),
                 user: str = get_sql_username(),
                 password: str = get_sql_password()) -> None:
        self.host = host
        self.schema = schema_name
        self.user = user
        self.password = password

        # Checking host suffix
        # TODO: move to sdk.get_sql_hostname
        if not (self.host.endswith("circ.zone") or self.host.endswith("circlez.ai")):
            logger.warning(
                f"Your RDS_HOSTNAME={self.host} which is not what is expected")
        self.connection: mysql.connector = None
        self._cursor: Cursor or None = None
        self._connect_to_db()
        connections_pool[schema_name] = self

    def reconnect(self) -> None:
        self.connection.reconnect()
        self._cursor = self.cursor()
        self.set_schema(self.schema)

    def _connect_to_db(self):
        self.connection = mysql.connector.connect(
            host=self.host,
            user=self.user,
            password=self.password,
            database=self.schema
        )
        self._cursor = self.cursor()
        self.set_schema(self.schema)

    def database_info(self):
        return f"host={self.host}, user={self.user}, schema={self.schema}"

    def close(self) -> None:

        try:
            if self._cursor:
                self._cursor.close()
                logger.info("Cursor closed successfully.")
        except Exception as exception:
            logger.error(f"connection.py close() {self.database_info()}", object={"exception": exception})

        if self.connection and self.connection.is_connected():
            self.connection.close()
            connections_pool.pop(self.schema, None)
            logger.info("Connection closed successfully.")

    def cursor(self, dictionary=False, buffered=False) -> Cursor:

        cursor_instance = Cursor(self.connection.cursor(
            dictionary=dictionary, buffered=buffered))

        return cursor_instance

    def commit(self) -> None:

        self.connection.commit()

    def set_schema(self, new_schema: str or None) -> None:
        if not new_schema:
            return
        if self.schema == new_schema:
            return

        if self._cursor and self.connection and self.connection.is_connected():
            use_query = f"USE `{new_schema}`;"
            self._cursor.execute(use_query)
            connections_pool[new_schema] = self
            connections_pool.pop(self.schema, None)
            self.schema = new_schema
            logger.info(f"Switched to schema: {new_schema}")
        else:
            raise Exception(
                "Connection is not established. The database will be used on the next connect.")

    def rollback(self):

        self.connection.rollback()

    def start_transaction(self):

        self.connection.start_transaction()



import os
import re
import sys
from typing import Any

import mysql.connector
from database_infrastructure_local.number_generator import NumberGenerator
from logger_local.MetaLogger import MetaLogger

from .connector import Connector
from .constants import DEFAULT_SQL_SELECT_LIMIT, LOGGER_CRUD_CODE_OBJECT
from .table_definition import table_definition
from .utils import (process_insert_data_json, process_update_data_json,
                    validate_none_select_table_name,
                    validate_select_table_name)


class GenericCRUD(metaclass=MetaLogger, object=LOGGER_CRUD_CODE_OBJECT):
    """A class that provides generic CRUD functionality.
    There are 4 main functions to create, read, update and delete data from the database.
    The rest of the functions are helper functions or wrappers around the main functions."""

    def __init__(self, *, default_schema_name: str,
                 default_table_name: str = None,
                 default_view_table_name: str = None,
                 default_id_column_name: str = None,
                 is_test_data: bool = False) -> None:
        """Initializes the GenericCRUD class. If a connection is not provided, a new connection will be created."""
        self.schema_name = default_schema_name
        self.connection = Connector.connect(schema_name=default_schema_name)
        self.cursor = self.connection.cursor()
        self.default_column = default_id_column_name
        self.default_table_name = default_table_name
        self.default_view_table_name = default_view_table_name or self._get_view_name(default_table_name)
        self.is_test_data = is_test_data

    def insert(self, *, schema_name: str = None, table_name: str = None, data_json: dict = None,
               ignore_duplicate: bool = False) -> int:
        """Inserts a new row into the table and returns the id of the new row or -1 if an error occurred."""

        if ignore_duplicate:
            self.logger.warning("GenericCRUD.insert using ignore_duplicate, is it really needed?")
        table_name = table_name or self.default_table_name
        schema_name = schema_name or self.schema_name
        self._validate_args(args=locals())

        if table_name in table_definition:
            if table_definition[table_name]["is_number_column"]:
                view_name = self._get_view_name(table_name)
                number = NumberGenerator.get_random_number(schema_name=schema_name, view_name=view_name)
                data_json["number"] = number
        else:
            self.logger.warning(f"database-mysql-local-python generic_crud.py Table {table_name} not found in "
                                f"database-mysql-local.table_definition_table data structure, we might need to run sql2code")
        # TODO: later we may want to add a check for the table_definition to see if the table is a test table
        # Then We will be able to remove the parameter is_test_data
        # and only call self.__is_test_data() without the following if/else
        if self.is_test_data:
            if self.__is_test_data():
                data_json["is_test_data"] = 1
            else:
                data_json["is_test_data"] = 0

        columns, values, data_json = process_insert_data_json(data_json=data_json)
        # We removed the IGNORE from the SQL Statement as we want to return the id of the existing row
        insert_query = "INSERT " + \
                       f"INTO `{schema_name}`.`{table_name}` ({columns}) " \
                       f"VALUES ({values});"
        try:
            self.cursor.execute(insert_query, tuple(data_json.values()))
            self.connection.commit()
            inserted_id = self.cursor.lastrowid()
        except mysql.connector.errors.IntegrityError as exception:
            if ignore_duplicate:
                self.logger.warning("GenericCRUD.insert: existing record found, selecting it's id")
                inserted_id = self._get_existing_duplicate_id(table_name, exception)
            else:
                raise exception

        return inserted_id

    def upsert(self, *, schema_name: str = None, table_name: str = None, data_json: dict = None) -> int:
        """Inserts a new row into the table, or updates an existing row if it already exists,
          and returns the id of the new row or -1 if an error occurred."""

        schema_name = schema_name or self.schema_name
        table_name = table_name or self.default_table_name
        self._validate_args(args=locals())

        if table_name in table_definition:
            if table_definition[table_name]["is_number_column"]:
                view_name = self._get_view_name(table_name)
                number = NumberGenerator.get_random_number(schema_name=schema_name, view_name=view_name)
                data_json["number"] = number
        else:
            self.logger.warning(f"database-mysql-local-python generic_crud.py Table {table_name} not found in "
                                f"database-mysql-local.table_definition_table data structure, we might need to run sql2code")
        # TODO: later we may want to add a check for the table_definition to see if the table is a test table
        if self.is_test_data:
            if self.__is_test_data():
                data_json["is_test_data"] = 1
            else:
                data_json["is_test_data"] = 0

        # TODO: process_insert_data_json should add created_user_id and updated_user_id ...
        columns, values, data_json = process_insert_data_json(data_json=data_json)
        # TODO: process_update_data_json should add updated_user_id and ... but not created_user_id
        set_values, data_json = process_update_data_json(data_json)
        insert_query = "INSERT " + \
                       f"INTO `{schema_name}`.`{table_name}` ({columns}) " \
                       f"VALUES ({values}) " \
                       f"ON DUPLICATE KEY UPDATE {set_values} updated_timestamp=CURRENT_TIMESTAMP();"
        self.cursor.execute(insert_query, tuple(data_json.values()) * 2)
        self.connection.commit()
        inserted_id = self.cursor.lastrowid()

        return inserted_id

    def _get_existing_duplicate_id(self, table_name: str, error: Exception) -> int:
        pattern = r'Duplicate entry \'(.+?)\' for key \'(.+?)\''
        match = re.search(pattern, str(error))
        if not match:  # a different error
            raise error
        duplicate_value = match.group(1)
        query = """
        SELECT COLUMN_NAME
        FROM INFORMATION_SCHEMA.KEY_COLUMN_USAGE
        WHERE TABLE_NAME = %s AND CONSTRAINT_NAME = "PRIMARY";
        """
        self.cursor.execute(query, (table_name,))
        column_name = self.cursor.fetchone()[0]
        if column_name:
            select_query = f"SELECT {column_name} FROM `{table_name}` WHERE {column_name} = %s LIMIT 1;"
            self.cursor.execute(select_query, (duplicate_value,))
            return self.cursor.fetchone()[0]
        else:  # Column name for constraint not found
            raise error

    def update_by_id(self, *, schema_name: str = None, table_name: str = None, id_column_name: str = None,
                     id_column_value: Any = None, data_json: dict = None,
                     limit: int = DEFAULT_SQL_SELECT_LIMIT, order_by: str = None) -> None:
        """Updates data in the table by ID."""

        table_name = table_name or self.default_table_name
        id_column_name = id_column_name or self.default_column

        if id_column_name:
            if id_column_value is None:
                where = f"`{id_column_name}` IS NULL"
                extra_sql_params = None
            else:
                where = f"`{id_column_name}`=%s"
                extra_sql_params = (id_column_value,)
            self.update_by_where(schema_name=schema_name, table_name=table_name, where=where,
                                 data_json=data_json, params=extra_sql_params,
                                 limit=limit, order_by=order_by)
        else:
            raise Exception("Update by id requires an id_column_name")

    def update_by_where(self, *, schema_name: str = None, table_name: str = None, where: str = None,
                        params: tuple = None, data_json: dict = None,
                        limit: int = DEFAULT_SQL_SELECT_LIMIT, order_by: str = None) -> None:
        """Updates data in the table by WHERE.
        Example:
        "UPDATE table_name SET A=A_val, B=B_val WHERE C=C_val AND D=D_val"
        translates into:
        update_by_where(table_name="table_name",
                        data_json={"A": A_val, "B": B_val},
                        where="C=%s AND D=%s",
                        params=(C_val, D_val)"""

        table_name = table_name or self.default_table_name
        schema_name = schema_name or self.schema_name
        self._validate_args(args=locals())

        set_values, data_json = process_update_data_json(data_json)
        if not where:
            raise Exception("update_by_where requires a 'where'")

        update_query = f"UPDATE `{schema_name}`.`{table_name}` " \
                       f"SET {set_values} updated_timestamp=CURRENT_TIMESTAMP() " \
                       f"WHERE {where} " + \
                       (f"ORDER BY {order_by} " if order_by else "") + \
                       f"LIMIT {limit};"
        params = params or tuple()
        self.cursor.execute(update_query, tuple(
            data_json.values()) + params)
        self.connection.commit()

    def delete_by_id(self, *, schema_name: str = None, table_name: str = None, id_column_name: str = None,
                     id_column_value: Any = None) -> None:
        """Deletes data from the table by id"""
        # logger, checks etc. are done inside delete_by_where
        id_column_name = id_column_name or self.default_column
        if id_column_name:  # id_column_value can be empty
            if id_column_value is None:
                where = f"`{id_column_name}` IS NULL"
                params = None
            else:
                where = f"`{id_column_name}`=%s"
                params = (id_column_value,)
            self.delete_by_where(schema_name=schema_name, table_name=table_name, where=where, params=params)
        else:
            raise Exception("Delete by id requires an id_column_name and id_column_value.")

    def delete_by_where(self, *, schema_name: str = None, table_name: str = None, where: str = None,
                        params: tuple = None) -> None:
        """Deletes data from the table by WHERE."""

        table_name = table_name or self.default_table_name
        schema_name = schema_name or self.schema_name
        self._validate_args(args=locals())
        if not where:
            raise Exception("delete_by_where requires a 'where'")

        update_query = f"UPDATE `{schema_name}`.`{table_name}` " \
                       f"SET end_timestamp=CURRENT_TIMESTAMP() " \
                       f"WHERE {where};"
        self.cursor.execute(update_query, params)
        self.connection.commit()

    # TODO: test distinct
    def select_one_tuple_by_id(self, *, schema_name: str = None, view_table_name: str = None,
                               select_clause_value: str = "*",
                               id_column_name: str = None,
                               id_column_value: Any = None,
                               distinct: bool = False, order_by: str = "") -> tuple:
        """Selects one row from the table by ID and returns it as a tuple."""
        result = self.select_multi_tuple_by_id(schema_name=schema_name,
                                               view_table_name=view_table_name,
                                               select_clause_value=select_clause_value,
                                               id_column_name=id_column_name,
                                               id_column_value=id_column_value,
                                               distinct=distinct,
                                               limit=1,
                                               order_by=order_by)
        if result:
            return result[0]
        else:
            return tuple()

    def select_one_dict_by_id(self, *, schema_name: str = None, view_table_name: str = None,
                              select_clause_value: str = "*",
                              id_column_name: str = None,
                              id_column_value: Any = None,
                              distinct: bool = False, order_by: str = "") -> dict:
        """Selects one row from the table by ID and returns it as a dictionary (column_name: value)"""
        result = self.select_one_tuple_by_id(schema_name=schema_name,
                                             view_table_name=view_table_name,
                                             select_clause_value=select_clause_value,
                                             id_column_name=id_column_name,
                                             id_column_value=id_column_value,
                                             distinct=distinct,
                                             order_by=order_by)
        return self.convert_to_dict(result, select_clause_value)

    def select_one_tuple_by_where(self, *, schema_name: str = None, view_table_name: str = None,
                                  select_clause_value: str = "*",
                                  where: str = None, params: tuple = None,
                                  distinct: bool = False, order_by: str = "") -> tuple:
        """Selects one row from the table based on a WHERE clause and returns it as a tuple."""
        result = self.select_multi_tuple_by_where(schema_name=schema_name,
                                                  view_table_name=view_table_name,
                                                  select_clause_value=select_clause_value,
                                                  where=where,
                                                  params=params,
                                                  distinct=distinct,
                                                  limit=1,
                                                  order_by=order_by)
        if result:
            return result[0]
        else:
            return tuple()

    def select_one_dict_by_where(self, *, schema_name: str = None, view_table_name: str = None,
                                 select_clause_value: str = "*",
                                 where: str = None, params: tuple = None,
                                 distinct: bool = False, order_by: str = "") -> dict:
        """Selects one row from the table based on a WHERE clause and returns it as a dictionary."""
        result = self.select_one_tuple_by_where(schema_name=schema_name,
                                                view_table_name=view_table_name,
                                                select_clause_value=select_clause_value,
                                                where=where,
                                                params=params,
                                                distinct=distinct,
                                                order_by=order_by)
        return self.convert_to_dict(result, select_clause_value)

    def select_multi_tuple_by_id(self, *, schema_name: str = None, view_table_name: str = None,
                                 select_clause_value: str = "*",
                                 id_column_name: str = None,
                                 id_column_value: Any = None,
                                 distinct: bool = False,
                                 limit: int = DEFAULT_SQL_SELECT_LIMIT, order_by: str = None) -> list:
        """Selects multiple rows from the table by ID and returns them as a
        list of tuples.
        send `id_column_name=''` if you want to select all rows and ignore default column"""
        id_column_name = id_column_name or self.default_column

        if not id_column_name:
            where = None
            params = None
        else:
            if id_column_value is None:
                where = f"{id_column_name} IS NULL"
                params = None
            else:
                where = f"{id_column_name}=%s"
                params = (id_column_value,)
        return self.select_multi_tuple_by_where(schema_name=schema_name,
                                                view_table_name=view_table_name,
                                                select_clause_value=select_clause_value,
                                                where=where,
                                                params=params,
                                                distinct=distinct,
                                                limit=limit,
                                                order_by=order_by)

    def select_multi_dict_by_id(
            self, *, schema_name: str = None, view_table_name: str = None, select_clause_value: str = "*",
            id_column_name: str = None, id_column_value: Any = None, distinct: bool = False,
            limit: int = DEFAULT_SQL_SELECT_LIMIT, order_by: str = None) -> list:
        """Selects multiple rows from the table by ID and returns them as a list of dictionaries."""
        result = self.select_multi_tuple_by_id(schema_name=schema_name,
                                               view_table_name=view_table_name,
                                               select_clause_value=select_clause_value,
                                               id_column_name=id_column_name,
                                               id_column_value=id_column_value,
                                               distinct=distinct,
                                               limit=limit,
                                               order_by=order_by)
        return [self.convert_to_dict(row, select_clause_value) for row in result]

    # Old name: select_multi_by_where
    def select_multi_tuple_by_where(self, *, schema_name: str = None, view_table_name: str = None,
                                    select_clause_value: str = "*",
                                    where: str = None, params: tuple = None, distinct: bool = False,
                                    limit: int = DEFAULT_SQL_SELECT_LIMIT,
                                    order_by: str = "") -> list:
        """Selects multiple rows from the table based on a WHERE clause and returns them as a list of tuples."""

        schema_name = schema_name or self.schema_name
        view_table_name = view_table_name or self.default_view_table_name
        self._validate_args(args=locals())

        # TODO: add ` to column names if they are not reserved words (like COUNT, ST_X(point), etc.)
        # select_clause_value = ",".join([f"`{x.strip()}`" for x in select_clause_value.split(",") if x != "*"])

        select_query = f"SELECT {'DISTINCT' if distinct else ''} {select_clause_value} " \
                       f"FROM `{schema_name}`.`{view_table_name}` " + \
                       (f"WHERE {where} " if where else "") + \
                       (f"ORDER BY {order_by} " if order_by else "") + \
                       f"LIMIT {limit};"
        self.cursor.execute(select_query, params)
        result = self.cursor.fetchall()

        return result

    def select_multi_dict_by_where(
            self, *, schema_name: str = None, view_table_name: str, select_clause_value: str = "*",
            where: str = None, params: tuple = None, distinct: bool = False,
            limit: int = DEFAULT_SQL_SELECT_LIMIT, order_by: str = None) -> list:
        """Selects multiple rows from the table based on a WHERE clause and returns them as a list of dictionaries."""
        result = self.select_multi_tuple_by_where(schema_name=schema_name,
                                                  view_table_name=view_table_name,
                                                  select_clause_value=select_clause_value,
                                                  where=where,
                                                  params=params,
                                                  distinct=distinct,
                                                  limit=limit,
                                                  order_by=order_by)
        return [self.convert_to_dict(row, select_clause_value) for row in result]

    # helper functions:
    def convert_to_dict(self, row: tuple, select_clause_value: str = "*") -> dict:
        """Returns a dictionary of the column names and their values."""
        if select_clause_value == "*":
            column_names = [col[0] for col in self.cursor.description()]
        else:
            column_names = [x.strip() for x in select_clause_value.split(",")]
        return dict(zip(column_names, row or tuple()))

    @staticmethod
    def _validate_args(args: dict) -> None:
        # args = locals() of the calling function
        required_args = ("data_json", "table_name", "view_table_name", "schema_name", "select_clause_value")
        for arg_name, arg_value in args.items():
            message = ""
            if arg_name in ("self", "__class__"):
                continue
            elif arg_name in required_args and not arg_value:
                message = f"Invalid value for {arg_name}: {arg_value}"
            elif arg_name == "table_name":
                validate_none_select_table_name(arg_value)
            elif arg_name == "view_table_name":
                validate_select_table_name(arg_value)

            # data_json values are allowed to contain ';', as we use them with %s (TODO: unless it's ToSQLInterface)
            if (arg_name == "data_json" and any(";" in str(x) for x in arg_value.keys()) or  # check columns
                    arg_name != "data_json" and ";" in str(arg_value)):
                message = f"Invalid value for {arg_name}: {arg_value} (contains ';')"

            if message:
                raise Exception(message)

    @staticmethod
    def __is_test_data() -> bool:
        """ Check if running from a Unit Test file. """
        file_name = os.path.basename(sys.argv[0])

        if file_name.startswith('test_') or file_name.endswith('_test.py'):
            return True
        else:
            return False

    def set_schema(self, schema_name: str or None):
        """Sets the given schema to be the default schema."""
        if not schema_name:
            return

        if self.schema_name != schema_name:
            self.connection.set_schema(schema_name)
            self.schema_name = schema_name

    def close(self) -> None:
        """Closes the connection to the database."""

        self.connection.close()

    def _log_error_message(self, message: str, sql_statement: str, schema_name: str) -> str:
        return (f"{message} - SQL statement: {sql_statement}. "
                f"(user={self.connection.user}, host={self.connection.host}, schema={schema_name})")

    @staticmethod
    def _get_view_name(table_name: str or None) -> str or None:
        if table_name:
            return re.sub(r'(_table)$', '_view', table_name)

    def get_test_entity_id(self, *, entity_name: str, insert_function: callable, insert_kwargs: dict = None,
                           entity_creator: callable = None, create_kwargs: dict = None,
                           schema_name: str = None, view_name: str = None) -> int:
        """
        1. Check if there's an entity with is `is_test_data=True`.
        2. If there is, return its id.
        3. If not, create a new entity with `is_test_data=True` and return its id.
        (assuming entity_creator expects `is_test_data` as parameters,
            and returns the expected argument for insert_function)

        Example: get_test_entity_id(entity_name='person', entity_creator=Person, insert_function=PersonsLocal.insert)
        """
        view_name = view_name or self.default_view_table_name
        select_clause_value = entity_name + "_id"
        fetched_result = self.select_one_dict_by_id(schema_name=schema_name or self.schema_name,
                                                    view_table_name=view_name,
                                                    id_column_name='is_test_data',
                                                    id_column_value='1',
                                                    select_clause_value=select_clause_value)
        if fetched_result:
            test_entity_id = fetched_result[select_clause_value]
        else:
            insert_kwargs = insert_kwargs or {}
            create_kwargs = create_kwargs or {}
            if entity_creator:
                entity_result = entity_creator(is_test_data=True, **create_kwargs)
                test_entity_id = insert_function(entity_result, **insert_kwargs)
            else:
                test_entity_id = insert_function(is_test_data=True, **insert_kwargs)
        return test_entity_id
